/*
             LUFA Library
     Copyright (C) Dean Camera, 2015.

  dean [at] fourwalledcubicle [dot] com
           www.lufa-lib.org
*/

/*
  Copyright 2015  Dean Camera (dean [at] fourwalledcubicle [dot] com)

  Permission to use, copy, modify, distribute, and sell this
  software and its documentation for any purpose is hereby granted
  without fee, provided that the above copyright notice appear in
  all copies and that both that the copyright notice and this
  permission notice and warranty disclaimer appear in supporting
  documentation, and that the name of the author not be used in
  advertising or publicity pertaining to distribution of the
  software without specific, written prior permission.

  The author disclaims all warranties with regard to this
  software, including all implied warranties of merchantability
  and fitness.  In no event shall the author be liable for any
  special, indirect or consequential damages or any damages
  whatsoever resulting from loss of use, data or profits, whether
  in an action of contract, negligence or other tortious action,
  arising out of or in connection with the use or performance of
  this software.
*/

/** \file
 *
 *  Main source file for the iUSB2SerialMux demo. This file contains the main tasks of the demo and
 *  is responsible for the initial application hardware configuration.
 */

#include "main.h"

/**************************************************************************************/

uint32_t EEMEM eep_baudrate;

uint32_t baudrate = 115200; // replacement for the UART_BAUDRATE definition

/**************************************************************************************/

/** Contains the current baud rate and other settings of the first virtual serial port. While this demo does not use
 *  the physical USART and thus does not use these settings, they must still be retained and returned to the host
 *  upon request or the host will assume the device is non-functional.
 *
 *  These values are set by the host via a class-specific request, however they are not required to be used accurately.
 *  It is possible to completely ignore these value or use other settings as the host is completely unaware of the physical
 *  serial link characteristics and instead sends and receives data in endpoint streams.
 */
static CDC_LineEncoding_t LineEncoding1 = { .BaudRateBPS = 0,
                                            .CharFormat  = CDC_LINEENCODING_OneStopBit,
                                            .ParityType  = CDC_PARITY_None,
                                            .DataBits    = 8                            };

/** Contains the current baud rate and other settings of the second virtual serial port. While this demo does not use
 *  the physical USART and thus does not use these settings, they must still be retained and returned to the host
 *  upon request or the host will assume the device is non-functional.
 *
 *  These values are set by the host via a class-specific request, however they are not required to be used accurately.
 *  It is possible to completely ignore these value or use other settings as the host is completely unaware of the physical
 *  serial link characteristics and instead sends and receives data in endpoint streams.
 */
static CDC_LineEncoding_t LineEncoding2 = { .BaudRateBPS = 0,
                                            .CharFormat  = CDC_LINEENCODING_OneStopBit,
                                            .ParityType  = CDC_PARITY_None,
                                            .DataBits    = 8                            };


/** Main program entry point. This routine configures the hardware required by the application, then
 *  enters a loop to run the application tasks in sequence.
 */
int main(void)
{
  SetupHardware();

  // initialize the command interpreter
  cc_init();

  SET_ERR_MASK(ERRMASK_USB_NOTREADY);

  GlobalInterruptEnable();

  for (;;)
  {
    CDC1_Task();
    CDC2_Task();

    USB_USBTask();

    //uart_putc('1');
  }
}

/** Configures the board hardware and chip peripherals for the demo's functionality. */
void SetupHardware(void)
{
#if (ARCH == ARCH_AVR8)
  /* Disable watchdog if enabled by bootloader/fuses */
  MCUSR &= ~(1 << WDRF);
  wdt_disable();

  /* Disable clock division */
  clock_prescale_set(clock_div_1);

#elif (ARCH == ARCH_XMEGA)
  /* Start the PLL to multiply the 2MHz RC oscillator to 32MHz and switch the CPU core to run from it */
  XMEGACLK_StartPLL(CLOCK_SRC_INT_RC2MHZ, 2000000, F_CPU);
  XMEGACLK_SetCPUClockSource(CLOCK_SRC_PLL);

  /* Start the 32MHz internal RC oscillator and start the DFLL to increase it to 48MHz using the USB SOF as a reference */
  XMEGACLK_StartInternalOscillator(CLOCK_SRC_INT_RC32MHZ);
  XMEGACLK_StartDFLL(CLOCK_SRC_INT_RC32MHZ, DFLL_REF_INT_USBSOF, F_USB);

  PMIC.CTRL = PMIC_LOLVLEN_bm | PMIC_MEDLVLEN_bm | PMIC_HILVLEN_bm;
#endif

  // set the three MUX control lines to output
  DDRB  |= (1 << PIN4) | (1 << PIN5) | (1 << PIN6);
  // set every pin to low
  EN_MUX_LINE0;

  DDRB |= (1 << PIN0) | (1 << PIN1);

  LED0_OFF;
  LED1_OFF;

  /* UART Hardware Initialization */
  eeprom_busy_wait();
  baudrate = eeprom_read_dword(&eep_baudrate);

  if (baudrate == 0xffffffff)
  {
    eeprom_busy_wait();
    baudrate = 115200;
    eeprom_write_dword(&eep_baudrate, baudrate);
  }

  uart_init( UART_BAUD_SELECT(baudrate, F_CPU) );
  sei();

  /* USB Hardware Initialization */
  USB_Init();
}

/** Event handler for the USB_Connect event. This indicates that the device is enumerating via the status LEDs and
 *  starts the library USB task to begin the enumeration and USB management process.
 */
void EVENT_USB_Device_Connect(void)
{
  /* Indicate USB enumerating */
  SET_ERR_MASK(ERRMASK_USB_ENUMERATING);
}

/** Event handler for the USB_Disconnect event. This indicates that the device is no longer connected to a host via
 *  the status LEDs and stops the USB management and CDC management tasks.
 */
void EVENT_USB_Device_Disconnect(void)
{
  /* Indicate USB not ready */
  SET_ERR_MASK(ERRMASK_USB_NOTREADY);
}

/** Event handler for the USB_ConfigurationChanged event. This is fired when the host set the current configuration
 *  of the USB device after enumeration - the device endpoints are configured and the CDC management tasks are started.
 */
void EVENT_USB_Device_ConfigurationChanged(void)
{
  bool ConfigSuccess = true;

  /* Setup first CDC Interface's Endpoints */
  ConfigSuccess &= Endpoint_ConfigureEndpoint(CDC1_TX_EPADDR, EP_TYPE_BULK, CDC_TXRX_EPSIZE, 1);
  ConfigSuccess &= Endpoint_ConfigureEndpoint(CDC1_RX_EPADDR, EP_TYPE_BULK, CDC_TXRX_EPSIZE, 1);
  ConfigSuccess &= Endpoint_ConfigureEndpoint(CDC1_NOTIFICATION_EPADDR, EP_TYPE_INTERRUPT, CDC_NOTIFICATION_EPSIZE, 1);

  /* Setup second CDC Interface's Endpoints */
  ConfigSuccess &= Endpoint_ConfigureEndpoint(CDC2_TX_EPADDR, EP_TYPE_BULK, CDC_TXRX_EPSIZE, 1);
  ConfigSuccess &= Endpoint_ConfigureEndpoint(CDC2_RX_EPADDR, EP_TYPE_BULK, CDC_TXRX_EPSIZE, 1);
  ConfigSuccess &= Endpoint_ConfigureEndpoint(CDC2_NOTIFICATION_EPADDR, EP_TYPE_INTERRUPT, CDC_NOTIFICATION_EPSIZE, 1);

  /* Reset line encoding baud rates so that the host knows to send new values */
  LineEncoding1.BaudRateBPS = 0;
  LineEncoding2.BaudRateBPS = 0;

  /* Indicate endpoint configuration success or failure */
  SET_ERR_MASK(ConfigSuccess ? ERRMASK_USB_READY : ERRMASK_USB_ERROR);
}

/** Event handler for the USB_ControlRequest event. This is used to catch and process control requests sent to
 *  the device from the USB host before passing along unhandled control requests to the library for processing
 *  internally.
 */
void EVENT_USB_Device_ControlRequest(void)
{
  /* Determine which interface's Line Coding data is being set from the wIndex parameter */
  void* LineEncodingData = (USB_ControlRequest.wIndex == 0) ? &LineEncoding1 : &LineEncoding2;

  /* Process CDC specific control requests */
  switch (USB_ControlRequest.bRequest)
  {
    case CDC_REQ_GetLineEncoding:
      if (USB_ControlRequest.bmRequestType == (REQDIR_DEVICETOHOST | REQTYPE_CLASS | REQREC_INTERFACE))
      {
        Endpoint_ClearSETUP();

        /* Write the line coding data to the control endpoint */
        Endpoint_Write_Control_Stream_LE(LineEncodingData, sizeof(CDC_LineEncoding_t));
        Endpoint_ClearOUT();
      }

      break;
    case CDC_REQ_SetLineEncoding:
      if (USB_ControlRequest.bmRequestType == (REQDIR_HOSTTODEVICE | REQTYPE_CLASS | REQREC_INTERFACE))
      {
        Endpoint_ClearSETUP();

        /* Read the line coding data in from the host into the global struct */
        Endpoint_Read_Control_Stream_LE(LineEncodingData, sizeof(CDC_LineEncoding_t));
        Endpoint_ClearIN();
      }

      break;
    case CDC_REQ_SetControlLineState:
      if (USB_ControlRequest.bmRequestType == (REQDIR_HOSTTODEVICE | REQTYPE_CLASS | REQREC_INTERFACE))
      {
        Endpoint_ClearSETUP();
        Endpoint_ClearStatusStage();
      }

      break;
  }
}

/** Function to manage CDC data transmission and reception to and from the host for the first CDC interface,
 *  which sends answers or response data to the host.
 */
void CDC1_Task(void)
{
  /* Device must be connected and configured for the task to run */
  if (USB_DeviceState != DEVICE_STATE_Configured)
    return;

  //===========================================================================
  /* Select the Serial Rx Endpoint */
  Endpoint_SelectEndpoint(CDC1_RX_EPADDR);

  if (Endpoint_IsOUTReceived())
  {
    /* Create a temp buffer big enough to hold the incoming endpoint packet */
    uint8_t  Buffer[Endpoint_BytesInEndpoint()];

    /* Remember how large the incoming packet is */
    uint16_t DataLength = Endpoint_BytesInEndpoint();

    /* Read in the incoming packet into the buffer */
    Endpoint_Read_Stream_LE(&Buffer, DataLength, NULL);

    /* Finalize the stream transfer to send the last packet */
    Endpoint_ClearOUT();

    for (uint16_t i = 0; i < DataLength; i++)
    {
      // process the received data and descide to do an action
      LED0_ON;
      cc_processData(Buffer[i]);
      LED0_OFF;
    }

  }

}

void USB_serialStreamWriteC(char *data, uint16_t len)
{
  //===========================================================================
  /* Determine if data/answeres should be sent to the host
   * the previous RX section should be clarify that behaviour.
   */

  /* Flag management - Only allow one string to be sent per action */
  if (data != NULL && len > 0 && LineEncoding1.BaudRateBPS)
  {

    LED1_ON;

    /* Select the Serial Tx Endpoint */
    Endpoint_SelectEndpoint(CDC1_TX_EPADDR);

    /* Write the String to the Endpoint */
    Endpoint_Write_Stream_LE(data, len, NULL);

    /* Finalize the stream transfer to send the last packet */
    Endpoint_ClearIN();

    /* Wait until the endpoint is ready for another packet */
    Endpoint_WaitUntilReady();

    /* Send an empty packet to ensure that the host does not buffer data sent to it */
    Endpoint_ClearIN();

    LED1_OFF;
  }

}

/** Function to manage CDC data transmission and reception to and from the host for the second CDC interface,
 *  which sends all data received from a node (mux) during USART to the host.
 */
void CDC2_Task(void)
{
  /* Device must be connected and configured for the task to run */
  if (USB_DeviceState != DEVICE_STATE_Configured)
    return;

  //===========================================================================
  /* Select the Serial Rx Endpoint */
  Endpoint_SelectEndpoint(CDC2_RX_EPADDR);

  /* Check to see if any data has been received */
  if (Endpoint_IsOUTReceived())
  {
    /* Create a temp buffer big enough to hold the incoming endpoint packet */
    uint8_t  Buffer[Endpoint_BytesInEndpoint()];

    /* Remember how large the incoming packet is */
    uint16_t DataLength = Endpoint_BytesInEndpoint();

    /* Read in the incoming packet into the buffer */
    Endpoint_Read_Stream_LE(&Buffer, DataLength, NULL);

    /* Finalize the stream transfer to send the last packet */
    Endpoint_ClearOUT();

    // TODO at this point send the data to the USART
    // Send USART &Buffer
    for (uint16_t i = 0; i < DataLength; i++)
    {
      uart_putc(Buffer[i]);
    }
  }

  //return;

  //===========================================================================
  uint8_t outBuffer[OUTPUT_BUFFER_SIZE];
  // TODO read the USART data and send them to the host
  // Fill &Buffer with USART data or send the USART input buffer direct
  uint16_t cnt = 0;
  int c = uart_getc();
  while (!(c & UART_NO_DATA) && cnt < OUTPUT_BUFFER_SIZE)
  {
    //LED0_ON;
    outBuffer[cnt] = c;
    c = uart_getc();
    cnt++;
  }

  /*
  cnt = 1;
  outBuffer[0] = '2';
  */

  // send the data which was received from the uart connection
  if (cnt > 0)
  {
    /* Select the Serial Tx Endpoint */
    Endpoint_SelectEndpoint(CDC2_TX_EPADDR);

    /* Write the received data to the endpoint */
    Endpoint_Write_Stream_LE(&outBuffer, cnt, NULL);

    /* Finalize the stream transfer to send the last packet */
    Endpoint_ClearIN();

    /* Wait until the endpoint is ready for the next packet */
    Endpoint_WaitUntilReady();

    /* Send an empty packet to prevent host buffering */
    Endpoint_ClearIN();
  }
}

